#include <math.h>
#include "export.h"
#include "players.h"

#define ZEROINIT(cls) memset(((char*)this)+4, 0, sizeof(cls)-4)

//////////////////////////////////////////////////////////////////////////

void EagerPlayer::InitPlayer()
{ 
	ZEROINIT(EagerPlayer);

	incr = 1000;
	decr = -900;
	up	= 1.3;
	down = .12;
	impact = 7;

	InitTable();
}

void EagerPlayer::RefreshBoard()
{
	InitPlayer();

	for (int i = 0; i < TABLE_SIZE; i++)
	for (int j = 0; j < TABLE_SIZE; j++)
	{
		Field f(i,j);

		int col = GetTableField(f);

		if (col)
		{
			ourcolor = col;
			SetMove(f);
			was_refresh = true;
		}
	}

	ourcolor = (char)NextColour();
}

void EagerPlayer::InitTable()
{
	for (ourcolor = black; ourcolor<= white; ourcolor++)
		for (int i = -1; i < TABLE_SIZE+1; i++)
			for (int j = -1; j < TABLE_SIZE+1; j++)
			{
				Field f(i,j);

				if (IsValid(f))
					continue;

				SetMove(f);
			}

	ourcolor = black;
}

//////////////////////////////////////////////////////////////////////////

void EagerPlayer::SetLastMove()
{
	if (nMove())
	{
		long last = Game(nMove());

		ourcolor = (char)GetTableField(last);
		SetMove(last);
	}
}

//////////////////////////////////////////////////////////////////////////

void EagerPlayer::RecalcSum(int us, int dir, const Field& f, double val)
{
	double oldval = Val(us, dir, f) + Val(us, OppDir(dir), f);

	Val(us, 8, f) -= ValPower(oldval);
	
	double newval = val + Val(us, OppDir(dir), f);

	Val(us, 8, f) += ValPower(newval);
}

//////////////////////////////////////////////////////////////////////////

void EagerPlayer::SetMove(const Field& fld)
{
	int us = ourcolor - 1;
	int dir;

	if (IsValid(fld))
	for (dir = 0; dir < 8; dir++)
	{
		double val = Val(us, dir, fld);

		val += incr;
		Val(us, dir, fld) = val;
		val *= up;

		Field f = fld;

		for (int offs = 1; offs < impact; offs++)
		{
			f.Step(dir, 1);

			if (!IsValid(f))
				break;

			if (Table(f))
			{
				if (Table(f) == ourcolor)
				{
					val += incr;
					Val(us, dir, f) = val;
					val *= up;
				}
				else
					 break;
			}
			else
			{
				RecalcSum(us, dir, f, val);
				Val(us, dir, f) = val;
				val *= down;
			}
		}
	}

	int them = !us;

	for (dir = 0; dir < 8; dir++)
	{
		double val = decr * up;

		Field f = fld;

		for (int offs = 1; offs < impact; offs++)
		{
			f.Step(dir, 1);

			if (!IsValid(f))
				break;

			if (Table(f))
			{
				if (Table(f) != ourcolor)
				{
					val += incr;
					Val(them, dir, f) = val;
					val *= up;
				}
				else
					 break;
			}
			else
			{
				RecalcSum(them, dir, f, val);
				Val(them, dir, f) = val;
				val *= down;
			}
		}
	}

	if (IsValid(fld))
	{
		Table(fld) = ourcolor; 
		ourcolor = (ourcolor == black ? white : black);
	}
}

/////////////////////////////////// virtuals /////////////////////////////
#pragma optimize( "", off )

long EagerPlayer::Move()
{
	if (!was_refresh)
	{
		SetLastMove();

		if (nMove() == 0)
		{
			Field f(7,7);
			
			ourcolor = black;
			SetMove(f);
			return f;
		}
	}

	was_refresh = false;

	Field fmax_us, fmax_them;
	double max_us=-1000000, max_them=-1000000;

	int us = ourcolor-1;

	for (int i = 0; i < TABLE_SIZE; i++)
	for (int j = 0; j < TABLE_SIZE; j++)
	{
		Field f(i,j);

		if (Table(f))
			continue;

		double val = Val(us, 8, f);

		if (val > max_us)
		{
			max_us = val;
			fmax_us = f;
		}

		val = Val(!us, 8, f);

		if (val > max_them)
		{
			max_them = val;
			fmax_them = f;
		}
	}

	Field fld = (max_us >= max_them ? fmax_us : fmax_them);

	SetMove(fld);

	return fld;
}

#pragma optimize( "", on )
//////////////////////////////////////////////////////////////////////////

double EagerPlayer::ValPower(double val)
{
	return val < 0 ? 0 : val * sqrt(sqrt(val));
}

//////////////////////////////////////////////////////////////////////////


